@@ -1,3 +1,4 @@
+from Acquisition import aq_parent
 from Products.Archetypes.interfaces import IReference
 from Products.Archetypes.Field import TextField
 from Products.Archetypes.exceptions import ReferenceException
@@ -25,8 +26,8 @@ def findObject(base, path):
 
 
 def getObjectsFromLinks(base, links):
-    """ determine actual objects (plus extra paths) refered to by given links """
-    objects = []
+    """ determine actual objects refered to by given links """
+    objects = set()
     url = base.absolute_url()
     scheme, host, path, query, frag = urlsplit(url)
     site = urlunsplit((scheme, host, '', '', ''))
@@ -35,31 +36,31 @@ def getObjectsFromLinks(base, links):
         if (not s and not h) or (s == scheme and h == host):    # relative or local url
             obj, extra = findObject(base, path)
             if obj:
-                extra += urlunsplit(('', '', '', q, f))
-                objects.append((obj, extra))
+                if IOFSImage.providedBy(obj):
+                    obj = aq_parent(obj)    # use atimage object for scaled images
+                objects.add(obj)
     return objects
 
 
 def modifiedArchetype(obj, event):
     """ an archetype based object was modified """
     try:    # TODO: is this a bug or a needed workaround?
-        for ref in obj.getReferences(relationship=referencedRelationship):
-            obj.deleteReference(ref)    # only remove forward references
+        existing = set(obj.getReferences(relationship=referencedRelationship))
     except AttributeError:
         return
-    refs = []
+    refs = set()
     for field in obj.Schema().fields():
         if isinstance(field, TextField):
             accessor = field.getAccessor(obj)
             links = extractLinks(accessor())
-            refs.extend(getObjectsFromLinks(obj, links))
-    for ref, extra in refs:
-        if IOFSImage.providedBy(ref):
-            ref = ref.aq_parent     # use atimage object for scaled images
+            refs = refs.union(getObjectsFromLinks(obj, links))
+    for ref in refs.difference(existing):   # add new references and...
         try:
             obj.addReference(ref, relationship=referencedRelationship)
         except ReferenceException:
             pass
+    for ref in existing.difference(refs):   # removed leftovers
+        obj.deleteReference(ref, relationship=referencedRelationship)
 
 
 def referenceRemoved(obj, event):
@@ -68,6 +69,11 @@ def referenceRemoved(obj, event):
     assert obj is event.object          # just making sure...
     if not obj.relationship == referencedRelationship:
         return                          # skip for other removed references
+    # if the object the event was fired on doesn't have a `REQUEST` attribute
+    # we can safely assume no direct user action was involved and therefore
+    # never raise a link integrity exception...
+    if not hasattr(obj, 'REQUEST'):
+        return
     storage = ILinkIntegrityInfo(obj.REQUEST)
     breaches = storage.getIntegrityBreaches()
     breaches.setdefault(obj.getTargetObject(), set()).add(obj.getSourceObject())
